#include "list.h"
#include <assert.h>
#include <stdlib.h>

struct d_list {
    d_list_it_t *top;
    d_list_it_t *bottom;
    int len;
    int ref;
};

#define check_list(list) \
    assert((list->len == 0 && !list->top && !list->bottom) || \
           (list->len > 0 && list->top && list->bottom))


d_list_t*
list_new(void)
{
    d_list_t *list = malloc(sizeof(d_list_t));
    list->top = NULL;
    list->bottom = NULL;
    list->ref = 1;
    list->len = 0;
    return list;
}

void
list_ref(d_list_t *list)
{
    ++list->ref;
}

void
list_unref(d_list_t *list)
{
    if (list && --list->ref == 0) {
        while (list->len)
            list_delete_link(list, list->top);
        free(list);
    }
}

d_list_it_t*
list_prepend(d_list_t *list, void *data)
{
    d_list_it_t *n = malloc(sizeof(d_list_it_t));
    n->data = data;

    n->prev = NULL;
    n->next = list->top;
    if (n->next) n->next->prev = n;

    list->top = n;
    if (!list->bottom) list->bottom = n;
    ++list->len;
    return n;
}

d_list_it_t*
list_append(d_list_t *list, void *data)
{
    d_list_it_t *n = malloc(sizeof(d_list_it_t));
    n->data = data;

    n->prev = list->bottom;
    n->next = NULL;
    if (n->prev) n->prev->next = n;

    list->bottom = n;
    if (!list->top) list->top = n;
    ++list->len;
    return n;
}

void
list_delete_link(d_list_t *list, d_list_it_t *pos)
{
    d_list_it_t *prev = pos->prev;
    d_list_it_t *next = pos->next;

    if (next)
        next->prev = prev;
    if (prev)
        prev->next = next;
    if (!next) {
        assert(list->bottom == pos);
        list->bottom = prev;
    }
    if (!prev) {
        assert(list->top == pos);
        list->top = next;
    }

    --list->len;
    check_list(list);
    free(pos);
}

void
list_move_before(d_list_t *list, d_list_it_t *move, d_list_it_t *before)
{
    d_list_it_t *prev, *next;

    /* these won't move it anywhere */
    if (move == before || move->next == before) return;

    prev = move->prev;
    next = move->next;

    /* remove it from the list */
    if (next) next->prev = prev;
    else      list->bottom = prev;
    if (prev) prev->next = next;
    else      list->top = next;

    /* reinsert it */
    if (before) {
        move->next = before;
        move->prev = before->prev;
        move->next->prev = move;
        if (move->prev) move->prev->next = move;
    }
    else {
        /* after the bottom */
        move->prev = list->bottom;
        move->next = NULL;
        if (move->prev) move->prev->next = move;
        list->bottom = move;
    }

    if (!move->prev) list->top = move;
}

void
list_remove(d_list_t *list, void *data)
{
    d_list_it_t *it = list_find(list, data);
    if (it) list_delete_link(list, it);
}

int
list_length(d_list_t *list)
{
    return list->len;
}

d_list_it_t*
list_top(d_list_t *list)
{
    return list->top;
}

d_list_it_t*
list_bottom(d_list_t *list)
{
    return list->bottom;
}

d_list_it_t*
list_nth(d_list_t *list, int n)
{
    int i;
    d_list_it_t *it;

    assert(n >= 0);
    assert(n < list->len);

    it = list->top;
    for (i = 0; i < n; ++i)
        it = it->next;
    return it;
}

d_list_it_t*
list_find(d_list_t *list, void *data)
{
    d_list_it_t *it;

    for (it = list->top; it; it = it->next)
        if (it->data == data) return it;
    return NULL;
}
